#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

from ._init import init_allocationmap, init_formats, init_logging
from .metadata import Metadata
from .mmusicc import MmusicC

name = "mmusicc"
